from xaal.schemas import devices
from xaal.lib import tools,Device
import pytuya
import time
import logging

# Tuya device aren't really network reliable, I get a bunch of connection reset
# This looks like they use a single socket for local network & cloud (SmartLife APP)
# connection. Everytime time I toggle a outlet in SmartLife I get a connection reset
# on the next update status. I tested w/ delays but not really working, so I switched
# to tenacity once again. 

#from tenacity import *
from tenacity import retry,stop_after_attempt


logger = logging.getLogger(__name__)
# disable pytuya debug
logging.getLogger('pytuya').setLevel(logging.INFO)

def now():
    return time.time()

class TuyaDev:
    def __init__(self,tuya_id,tuya_key,tuya_ip,base_addr):
        self.tuya_id  = tuya_id
        self.tuya_key = tuya_key
        self.tuya_ip  = tuya_ip
        self.tuya_dev = None
        self.base_addr = base_addr
        self.last_update = 0
        self.devices = []
        self.setup()
        self.init_properties()

    def setup(self):
        logger.warning('Please override setup()')

    def init_properties(self):
        for dev in self.devices:
            dev.vendor_id = 'IHSEV / Tuya'
            dev.hw_id = self.tuya_id
            dev.info = 'Tuya device: %s @ %s' % (self.tuya_id,self.tuya_ip)
            if len(self.devices) > 1:
                dev.group_id = self.base_addr + 'ff'
        
    def update_status(self):
        try:
            self._update_status()
        except RetryError as e:
            logger.warn(e)

    @retry(stop=stop_after_attempt(2))
    def _update_status(self):
        time.sleep(0.1)
        logger.info("Updating %s" % self.tuya_id)
        self.last_update = now()
        status = self.tuya_dev.status()
        logger.warning(status)
        self.on_status(status)

    def on_status(self,status):
        logger.warning('Please override on_status')
            
                
class OnOffMixing(object):

    def setup_onoff(self,tuya_dev,xaal_dev):
        tuya_dev.set_version(3.3)        
        self.tuya_dev = tuya_dev
        xaal_dev.methods['on']  = self.on
        xaal_dev.methods['off'] = self.off
        xaal_dev.methods['toggle'] = self.toggle
        self.xaal_dev = xaal_dev

    @retry(stop=stop_after_attempt(2))
    def on(self):
        self.tuya_dev.turn_on()
        self.update_status()

    @retry(stop=stop_after_attempt(2))
    def off(self):
        self.tuya_dev.turn_off()
        self.update_status()

    def toggle(self):
        power = self.xaal_dev.attributes['power']
        if power == None:
            self.update_status()
            power = self.xaal_dev.attributes['power']
        if power == True:
            self.off()
        elif power == False:
            self.on()

class PowerRelay(TuyaDev,OnOffMixing):
    def setup(self):
        relay = devices.powerrelay_toggle(self.base_addr + '00')
        relay.product_id = 'Generic Tuya Outlet'
        outlet = pytuya.OutletDevice(self.tuya_id,self.tuya_ip,self.tuya_key)
        outlet.set_version(3.3)
        self.setup_onoff(outlet,relay)
        self.devices.append(relay)

    def on_status(self,status):
        self.devices[0].attributes['power'] = status['dps'][1]
        

class SmartPlug(TuyaDev,OnOffMixing):
    def setup(self):
        # TBD switch to the right schema, but right now nothing exists
        relay = devices.powerrelay_toggle(self.base_addr + '00')
        relay.product_id = 'Tuya Smart Plug Outlet'
        plug = pytuya.OutletDevice(self.tuya_id,self.tuya_ip,self.tuya_key)
        self.setup_onoff(plug,relay)

        pmeter = devices.powermeter(self.base_addr + '01')
        pmeter.product_id = 'Tuya Smart Plug Power Meter'
        pmeter.new_attribute('voltage')
        pmeter.new_attribute('current')
        pmeter.del_attribute(pmeter.get_attribute('energy'))
        pmeter.del_attribute(pmeter.get_attribute('devices'))
        pmeter.unsupported_attributes = ['energy','devices']
        
        self.devices.append(relay)
        self.devices.append(pmeter)
        
    def on_status(self,status):
        dps = status['dps']
        print(dps)
        self.devices[0].attributes['power'] = dps.get('1',None)
        self.devices[1].attributes['current'] = int(dps.get('4',0)) / 1000
        self.devices[1].attributes['power'] = int(dps.get('5',0)) / 10
        self.devices[1].attributes['voltage'] = int(dps.get('6',0)) / 10
        
        
        
# TODO I don't have this kind of devices right now
class RGBLamp(TuyaDev):pass
class PowerRelayMetering(TuyaDev):pass
